@@ -55,6 +55,23 @@ impl Default for DisplayOptions {
     }
 }
 
+#[derive(Debug, Clone)]
+pub struct DiffOptions {
+    pub graph_limit: usize,
+    pub byte_limit: usize,
+    pub check_only: bool,
+}
+
+impl Default for DiffOptions {
+    fn default() -> Self {
+        Self {
+            graph_limit: DEFAULT_GRAPH_LIMIT,
+            byte_limit: DEFAULT_BYTE_LIMIT,
+            check_only: false,
+        }
+    }
+}
+
 fn app() -> clap::Command<'static> {
     Command::new("Difftastic")
         .override_usage(USAGE)
@@ -225,8 +242,15 @@ pub enum FileArgument {
     DevNull,
 }
 
+fn try_canonicalize(path: &Path) -> PathBuf {
+    path.canonicalize().unwrap_or_else(|_| path.into())
+}
+
 fn relative_to_current(path: &Path) -> PathBuf {
     if let Ok(current_path) = std::env::current_dir() {
+        let path = try_canonicalize(path);
+        let current_path = try_canonicalize(&current_path);
+
         if let Ok(rel_path) = path.strip_prefix(current_path) {
             return rel_path.into();
         }
@@ -265,12 +289,10 @@ impl FileArgument {
 
 pub enum Mode {
     Diff {
-        graph_limit: usize,
-        byte_limit: usize,
+        diff_options: DiffOptions,
         display_options: DisplayOptions,
         missing_as_empty: bool,
         set_exit_code: bool,
-        check_only: bool,
         language_override: Option<guess_language::Language>,
         /// The path where we can read the LHS file. This is often a
         /// temporary file generated by source control.
@@ -479,13 +501,17 @@ pub fn parse_args() -> Mode {
         in_vcs,
     };
 
-    Mode::Diff {
+    let diff_options = DiffOptions {
         graph_limit,
         byte_limit,
+        check_only,
+    };
+
+    Mode::Diff {
+        diff_options,
         display_options,
         missing_as_empty,
         set_exit_code,
-        check_only,
         language_override,
         lhs_path,
         rhs_path,
