@@ -246,7 +246,7 @@ impl<'a> Syntax<'a> {
     ) -> &'a Syntax<'a> {
         // If a parser hasn't cleaned up \r on CRLF files with
         // comments, discard it.
-        if content.ends_with("\r") {
+        if content.ends_with('\r') {
             content = &content[..content.len() - 1];
         }
 
@@ -583,6 +583,7 @@ pub enum AtomKind {
     Type,
     Comment,
     Keyword,
+    TreeSitterError,
 }
 
 /// Unlike atoms, tokens can be delimiters like `{`.
