@@ -53,6 +53,7 @@ extern "C" {
     fn tree_sitter_elixir() -> ts::Language;
     fn tree_sitter_elm() -> ts::Language;
     fn tree_sitter_elvish() -> ts::Language;
+    fn tree_sitter_erlang() -> ts::Language;
     fn tree_sitter_gleam() -> ts::Language;
     fn tree_sitter_go() -> ts::Language;
     fn tree_sitter_hare() -> ts::Language;
@@ -285,6 +286,19 @@ pub fn from_language(language: guess::Language) -> TreeSitterConfig {
                 .unwrap(),
             }
         }
+        Erlang => {
+            let language = unsafe { tree_sitter_erlang() };
+            TreeSitterConfig {
+                language,
+                atom_nodes: [].into(),
+                delimiter_tokens: vec![("(", ")"), ("{", "}"), ("[", "]")],
+                highlight_query: ts::Query::new(
+                    language,
+                    include_str!("../../vendor/highlights/erlang.scm"),
+                )
+                .unwrap(),
+            }
+        }
         Gleam => {
             let language = unsafe { tree_sitter_gleam() };
             TreeSitterConfig {
